import struct
import socket
from typing import Tuple, Union
from libscrc import ibm as crc16

from onbon.defaults import Devices, Errors, RESERVED, DST_ADDR, SRC_ADDR, PROTO
from onbon.dynamic_area.commands import (
    UpdateDynamicArea,
    DeleteDynamicArea,
    Ping,
    Command
)
from onbon.dynamic_area.utils import PacketEncoder


class PacketHeader:
    def __init__(
            self,
            device: int,
            dstaddr: int = DST_ADDR,
            srcaddr: int = SRC_ADDR,
            proto: int = PROTO,
            **kwargs
    ):
        """
        Creates packet header.

        Receives device code as argument. See :obj:`~defaults.Devices`
        """
        self.dstAddr = dstaddr
        self.srcAddr = srcaddr
        self.proto = proto
        self.device = Devices.get(device)

    @staticmethod
    def _check_crc16(crc: bytes, buffer: bytes, crc_byteorder='little'):
        return int.from_bytes(crc, crc_byteorder) == crc16(buffer)

    def to_bytes(self):
        return struct.pack(
            '<2H2BHI',
            self.dstAddr,
            self.srcAddr,
            self.proto,
            RESERVED,
            self.device,
            RESERVED
        )

    def _unpack(self, buffer: bytes):
        (
            self.dstAddr,
            self.srcAddr,
            self.proto,
            _,
            self.device
        ) = struct.unpack(
            '<2H2BH',
            buffer[:8]
        )


class Response(PacketHeader):
    def __init__(self, response: bytes, raise_on_corruption: bool = True):
        """
        Unpacks controllers response.

        If `raise_on_corruption` parameter is True it raises ValueError
        on crc16 mismatch.
        """
        super().__init__(Devices.NO_DEVICE)
        self.dataLen = None
        self.status = None
        self.error = None
        self.resData = None
        self.crc16 = None
        self.crcValid = False
        self._unpack(response)
        self.check_crc16(response, raise_on_corruption=raise_on_corruption)

    def __repr__(self):
        return '<CRC: %s; Status: %s; Error: %s>' % (
            'OK' if self.crcValid else 'CORRUPTED',
            'NACK' if self.status else 'ACK',
            Errors.get_key(self.error, 'UNKNOWN')
        )

    def _unpack(self, buffer: bytes):
        # currently just unpacks status, error, crc and raw response data
        super()._unpack(buffer)
        (
            self.status,
            self.error,
        ) = struct.unpack(
            '<2H',
            buffer[19:23]
        )
        self.resData = buffer[25:-2]
        self.crc16 = buffer[-2:]

    def check_crc16(self, buffer: bytes, raise_on_corruption):
        self.crcValid = self._check_crc16(self.crc16, buffer[:-2])
        if raise_on_corruption and not self.crcValid:
            raise ValueError('Response CRC16 mismatch error.')


class Request(PacketHeader):
    def __init__(
            self,
            address: Tuple[str, int],
            device: Union[str, int],
            command: Command,
            **kwargs
    ):
        """

        :param address: controller address as tuple (IP, PORT)
        :param device: string or integer representation of device
        :param command: command subclass instance, e.g. .commands.Ping
        """
        super().__init__(device, **kwargs)
        self.address = address
        self.command = command

    def to_bytes(self):
        header = super().to_bytes()
        command_data = self.command.to_bytes()
        data_len = len(command_data)
        packet = header + data_len.to_bytes(4, 'little') + command_data
        crc = crc16(packet).to_bytes(2, 'little')
        return PacketEncoder.frame_packet(packet + crc)

    def execute(self, timeout: float = 5) -> Response:
        """
        Sends command packet to controller
        and returns :class:`Response` instance.

        :param timeout: socket timeout
        """
        with socket.create_connection(
                self.address,
                timeout=timeout
        ) as sock:
            sock.sendall(self.to_bytes())
            response_bytes = sock.recv(1024)
            response = PacketEncoder.decode(response_bytes)
            return Response(response)


class LEDClient:
    """Onbon's 6th series TCP client"""
    def __init__(
            self,
            address: Tuple[str, int],
            device: Union[str, int]
    ):
        """

        :param address: address in (IP, PORT) format
        :param device: string or integer representation of device
        """
        self.address = address
        self.device = Devices.get(device)

    def ping(self) -> Response:
        """
        Sends a ping command to controller.

        :return: controllers response
        """
        return Request(
            self.address,
            self.device,
            Ping()
        ).execute(timeout=1)

    def delete_areas(self, *areaids) -> Response:
        """
        Receives dynamic area ids to delete.

        If nothing passed as argument it deletes all existing dynamic ares.

        :return: controllers response
        """
        return Request(
            self.address,
            self.device,
            DeleteDynamicArea(*areaids)
        ).execute()

    def update_areas(self, *dynamicareas) -> Response:
        """
        Packs given :class:`~.lib.DynamicArea` instances into one update
        command and sends it to controller.
        """
        if not dynamicareas:
            raise ValueError(
                'At least one dynamic area must be passed as argument.'
            )
        return Request(
            self.address,
            self.device,
            UpdateDynamicArea(*dynamicareas)
        ).execute()
