import struct

from onbon.defaults import RESERVED, Commands
from onbon.dynamic_area.lib import DynamicArea


class Command:
    """Base command class"""
    def __init__(self, command):
        self.retreq, self.cmdgroup, self.cmd = command

    def __bytes__(self):
        return self.to_bytes()

    def to_bytes(self):
        raise NotImplementedError


class Ping(Command):
    def __init__(self):
        super().__init__(Commands.PING)

    def to_bytes(self):
        return struct.pack(
            '<3BH',
            self.retreq,
            self.cmdgroup,
            self.cmd,
            RESERVED,
        )


class DeleteDynamicArea(Command):
    def __init__(self, *areaids: int):
        """
        Receives dynamic area ids to delete.

        If nothing passed as argument it deletes all existing dynamic ares.
        """
        super().__init__(Commands.DELETE)
        self.areanum = len(areaids)
        self.areaids = areaids

    def to_bytes(self):
        return struct.pack(
            '<3BHB%dB' % self.areanum,
            self.retreq,
            self.cmdgroup,
            self.cmd,
            RESERVED,
            self.areanum or 0xFF,
            *self.areaids
        )


class UpdateDynamicArea(Command):
    def __init__(self, *dynamicareas: DynamicArea):
        """
        Sends/updates given dynamic areas to controller.
        """
        super().__init__(Commands.UPDATE)
        self.dynamicareas = list(dynamicareas)
        self.area_ids = {area.id for area in self.dynamicareas}

    def __getitem__(self, i):
        return self.dynamicareas[i]

    def __delitem__(self, i):
        del self.dynamicareas[i]

    def pixel_data(self):
        return b''.join(area.pixel_data() for area in self.dynamicareas)

    def to_bytes(self):
        area_data = b''
        offset = 0
        for i, area in enumerate(self.dynamicareas):
            area_id = area.id or i
            while area_id in self.area_ids:
                area_id += 1
            buffer = area.area_data(start_offset=offset)
            area_data += struct.pack('<IB', len(buffer) + 1, area_id) + buffer
            offset += len(area.data)
        area_num = len(self.dynamicareas)
        return (
                struct.pack(
                    '<3B2H',
                    self.retreq,
                    self.cmdgroup,
                    self.cmd,
                    RESERVED,
                    area_num
                ) +
                area_data +
                self.pixel_data()
        )