import struct
from typing import Tuple
from PIL import (
    Image, ImageFont, ImageDraw,
    ImageOps, ImageColor
)

from onbon.defaults import RESERVED, RunModes, DisplayModes
from onbon.dynamic_area.utils import calc_padding


class ImageData:
    """
    A helper class that transforms image data into
    a format that the controller understands,
    e.g. pagination and converting to RGB888 mode.
    """
    def __init__(
            self,
            image: Image.Image,
            size: Tuple[int, int],
            action: str = None,
            align: str = 'top-left',
            xy: Tuple[int, int] = None,
            fit: bool = False,
            flip: bool = False,
            mirror: bool = False,
            **kwargs
    ):
        """
        Creates a paged dynamic area object from given Pillow.Image instance.

        Valid `action` parameter values:
            * resize - resizes imageto given area size [1]_ .
                BICUBIC interpolation is used by default.
                `resample` и `box` can be passed as kwargs to change
                interpolation method and box
                (see :meth:`PIL.Image.resize`).

            * paste - pastes given image into blank image of given
                area size [1]_ .
                Pasting position is regulated by `align` [2]_ or `xy` parameters.
                If `xy` parameter is given `align` parameter is ignored.

            * crop - crops image of area size from given image [1]_ .
                `xy` parameter can be passed to set upper-left coordinate of
                cropped image [(0, 0) used by deafult].

        .. [1] if parameter `fit` = True image will be resized to one page,
            otherwise it will be expanded to needed amount of pages.
        .. [2] see :py:meth:`.utils.calc_padding`

        :param image: Pillow image
        :param size: dynamic area size
        :param action: performed action if sizes mismatch
        :param align: positioning param
        :param xy: coordinate of pasting or cropping image
        :param fit: fit to given size flag
        :param flip: flips image vertically (up -> down)
        :param mirror: mirrors image horizontally (left -> right)
        :param kwargs: extra parameters
        """
        self.font_color = ImageColor.getrgb(
            kwargs.get('font_color', 'white')
        )
        self.bg_color = ImageColor.getrgb(
            kwargs.get('bg_color', 'black')
        )
        if image.mode != 'RGB':
            image = image.convert('RGB')
        if flip:
            image = ImageOps.flip(image)
        if mirror:
            image = ImageOps.mirror(image)
        w, h = size
        iw, ih = image.size
        if size != image.size:
            new_size = size if fit else (iw, h)
            if action is not None:
                if action == 'resize':
                    image = image.resize(
                        new_size,
                        resample=kwargs.get('resample', Image.BICUBIC),
                        box=kwargs.get('box')
                    )
                elif action == 'paste':
                    if not xy:
                        xy = calc_padding(new_size, image.size, align)
                    xy_2 = map(sum, zip(xy, image.size))
                    im = Image.new('RGB', new_size, self.bg_color)
                    im.paste(image, (*xy, *xy_2))
                    image = im
                elif action == 'crop':
                    if not xy:
                        xy = (0, 0)
                    xy_2 = map(sum, zip(xy, new_size))
                    image = image.crop((*xy, *xy_2))
                else:
                    raise ValueError(
                        'Area and image sizes are not equal. '
                        'Pass `action` parameter to properly transform image.'
                    )
            if not fit:
                pages = iw // w + bool(iw % w)
                xy = calc_padding((w * pages, h), image.size, align)
                xy_2 = map(sum, zip(xy, image.size))
                im = Image.new('RGB', (w * pages, h), self.bg_color)
                im.paste(image, (*xy, *xy_2))
                image = im
        self.size = size
        self._image = image

    def __bool__(self):
        return all(self._image.size)

    def __add__(self, other: 'ImageData'):
        if self.size[1] != other.size[1]:
            raise ValueError('Area heights of operands are not equal.')
        new_w = self._image.size[0] + other._image.size[0]
        im1 = self._image
        im2 = other._image
        im = Image.new('RGB', (new_w, self.size[1]))
        im.paste(im1, (0, 0, *im1.size))
        im.paste(im2, (im1.size[0], 0, *im.size))
        return ImageData(im, (new_w, self.size[1]))

    def __bytes__(self):
        w, h = self.size
        return b''.join(
            self._image.crop((w * i, 0, w * (i + 1), h)).tobytes()
            for i in range(self.page_count)
        )

    def __len__(self):
        w, h = self._image.size  # image size
        return w * h * 3  # RGB888 - 3 bytes per pixel

    @property
    def page_offset(self) -> int:
        """Bytes length of one page."""
        w, h = self.size  # dyn. area size
        return w * h * 3  # RGB888 - 3 bytes per pixel

    @property
    def page_count(self) -> int:
        """Dynamic area pages count"""
        return len(self) // self.page_offset

    def get_data(self):
        """
        Image bytes in specific cropped format.

        It crops image into pages and concatenates resulted pages bytes.
        """
        return bytes(self)

    def show(self, *args, **kwargs):
        """Shows image on the screen."""
        if self:
            self._image.show(*args, **kwargs)

    @classmethod
    def from_text(
            cls,
            text: str,
            size: Tuple[int, int],
            font: ImageFont.FreeTypeFont,
            bg_color='black',
            font_color='white',
            align='left',
            **kwargs
    ) -> 'ImageData':
        """
        Creates image from given text.

        If text contains newline chars then method
        :py:meth:`~ImageData.from_multiline_text` will be called instead.

        :param text: text
        :param size: dyn. area size
        :param font: PIL.ImageFont instance
        :param bg_color: background color
        :param font_color: text color
        :param align: text positioning
        :param kwargs: extra params
        """
        if '\n' in text:
            text = '\n'.join(
                map(lambda x: '%s ' % x, text.split('\n'))
            )  # prevents PIL eating half of last symbol on each line
            return cls.from_multiline_text(
                text, size, font,
                bg_color=bg_color,
                font_color=font_color,
                text_position=align,
                **kwargs
            )
        text = ' %s ' % text  # prevents PIL eating half of last symbol
        mask = font.getmask(
            text,
            mode='1',
            features=kwargs.get('features'),
            stroke_width=kwargs.get('stroke_width', 0)
        )
        w, h = size  # area size
        mw, mh = mask.size  # text mask size
        if mh > h and 'action' not in kwargs:
            raise ValueError(
                'Font height is bigger than dyn. area height.'
                'Reduce font size or pass `action` parameter'
                '(see :class:`ImageData` constructor docs).'
            )
        if mw > w:
            ew = w * (mw // w + bool(mw % w))  # extended width
            size = ew, h
        im = Image.new('RGB', size, bg_color)
        if all(mask.size):
            im_text = Image.new('RGB', mask.size, font_color)
            mask_image = Image.frombytes('L', mask.size, bytes(mask))
            padding = calc_padding(size, mask.size, align)
            im.paste(im_text, padding, mask_image)
        return cls(im, size=(w, h), bg_color=bg_color, font_color=font_color, **kwargs)

    @classmethod
    def from_multiline_text(
            cls,
            text: str,
            size: Tuple[int, int],
            font: ImageFont.FreeTypeFont,
            bg_color='black',
            font_color='yellow',
            text_position='center',
            text_align='center',
            spacing=2,
            **kwargs
    ) -> 'ImageData':
        """
        Creates image from multiline text.

        :param text: text
        :param size: dynamic area size
        :param font: PIL.ImageFont instance
        :param bg_color: background color
        :param font_color: text color
        :param text_position: text positioning
        :param text_align: text aligning (left, center, right)
        :param spacing: spacing between lines
        :param kwargs: extra params
        """
        im = Image.new('RGB', size, bg_color)
        draw = ImageDraw.Draw(im)
        draw.fontmode = '1'
        text_size = draw.textsize(text, font=font, spacing=spacing)
        padding = calc_padding(size, text_size, align=text_position)
        draw.text(padding, text, font=font, fill=font_color, align=text_align, spacing=spacing)
        # im.paste(text_im, (0, 0, *text_im.size), mask)
        return cls(im, size=size, bg_color=bg_color, font_color=font_color, **kwargs)

    @classmethod
    def isochromatic(
            cls,
            size: Tuple[int, int],
            color='black'
    ) -> 'ImageData':
        """
        Creates isochromatic image of given size and color.

        :param size: dynamic area size
        :param color: background color
        """
        im = Image.new('RGB', size, color)
        return cls(im, size=size, bg_color=color)

    @classmethod
    def empty(cls, size: Tuple[int, int]) -> 'ImageData':
        """
        Creates empty image with zero width and corresponding height.

        Can be used as initial value in `sum` method for concatenating
        ImageData instances of same height:
            >>> ims = [ImageData.isochromatic((64,32), 'red'), ImageData.isochromatic((32,32), 'lime')]
            >>> im = sum(ims, ImageData.empty((32,32)))

        :param size: dynamic area size
        """
        w, h = size
        im = Image.new('RGB', (0, h))
        return cls(im, size=size)

    @classmethod
    def hconcat(cls, size: Tuple[int, int], *image_datas: 'ImageData',
                align='center') -> 'ImageData':
        """
        Horizontal concatenation of given ImageData object into one of given size.
        Concatenates in same order as passed.
        All ImageData heights must less or equal to area's height,
        otherwise it raises ValueError.

        :param size: dynamic area size
        :param image_datas: concatenating objects
        :param align: alignment while concatenating (top, center, bottom)
        """
        w, h = size
        if any(h < im.size[1] for im in image_datas):
            raise ValueError('Heights mismatch!')
        ow = sum(im.size[0] for im in image_datas)  # overall
        concatenated = Image.new('RGB', (ow, h))
        x = 0
        for im in image_datas:
            iw, ih = im.size
            px, py = calc_padding((iw, h), (iw, ih), align=align)
            concatenated.paste(im._image, (px + x, py, px + x + iw, py + ih))
            x += iw
        return cls(concatenated, size)

    @classmethod
    def vconcat(cls, size: Tuple[int, int], *image_datas: 'ImageData',
                align='center') -> 'ImageData':
        """
        Vertical concatenation of given ImageData object into one of given size.
        Concatenates in same order as passed.
        All ImageData widths must less or equal to area's width,
        otherwise it raises ValueError.

        :param size: dynamic area size
        :param image_datas: concatenating objects
        :param align: alignment while concatenating (left, center, right)
        """
        w, h = size
        if any(w < im.size[0] for im in image_datas):
            raise ValueError('Widths mismatch!')
        oh = sum(im.size[1] for im in image_datas)  # overall
        concatenated = Image.new('RGB', (w, oh))
        y = 0
        for im in image_datas:
            iw, ih = im.size
            px, py = calc_padding((w, ih), (iw, ih), align=align)
            concatenated.paste(im._image, (px, py + y, px + iw, py + y + ih))
            y += ih
        return cls(concatenated, size)


class AreaFrame:
    def __init__(self):
        pass


class DynamicArea:
    def __init__(self, data: ImageData, **kwargs):
        """
        Valid dynamic area params and default values:
            * id = None - dynamic area id
                          If param is not passed it will be given automatically
                          at packet framing.
            * runmode = RunModes.CYCLE - dynamic area runmode (:obj:`~defaults.RunModes`)
            * timeout = 5 - dynamic area play time in seconds (0-65535)
            * relateallpro = True - attach to all programms flag
            * relateproserials = [0] -  list of programms' ids to attach.
                                        Ignored if `relateallpro` = True.
            * overlap = True - if True attached async programm stops,
                               otherwise tries to play it in parallel.
            * x = 0 - `x` coord of upper-left corner of dyn. area on the led screen
            * y = 0 - `y` coord of upper-left corner of dyn. area on the led screen
            * displaymode = DisplayModes.STATIC - page 'flipping' effect (:obj:`~defaults.RunModes`)
            * speed = 5 - page move speed (0-255: less is faster)
            * staytime = 0 - page staytime on the screen (unit is 10ms -> 100 = 1 sec.)
            * repeattime = 1 - repeat time cycles

        :param data: ImageData object
        :param kwargs: dynamic area params
        """
        self.data = data
        self.id = kwargs.get('id', None)
        self.runmode = kwargs.get('runmode', RunModes.CYCLE)
        self.timeout = kwargs.get('timeout', 5)  # in seconds
        self.relateallpro = kwargs.get('relateallpro', True)
        self.relateproserials = kwargs.get('relateproserials', [0])
        self.relatepronum = len(self.relateproserials)
        self.overlap = kwargs.get('overlap', True)
        self.areatype = 0x10
        self.x = kwargs.get('x', 0)
        self.y = kwargs.get('y', 0)
        self.width, self.height = self.data.size
        self.areaframe = kwargs.get('areaframe', 0)  # TODO: implement AreaFrame
        self.pagenum = self.data.page_count
        self.pagedatalen = 26  # 26 bytes kwargs.get('pagedatalen', 26)
        self.pagestyle = 0
        self.displaymode = kwargs.get('displaymode', DisplayModes.STATIC)
        self.clearmode = kwargs.get('clearmode', 0)
        self.speed = kwargs.get('speed', 0x05)  # меньше = быстрее
        self.staytime = kwargs.get('staytime', 0)  # 10ms/1
        self.repeattime = kwargs.get('repeattime', 1)
        self.validlen = kwargs.get('validlen', self.width)
        self.dataoffset = kwargs.get('datalen', self.data.page_offset)

    def update(self, **kwargs):
        """
        Updates dynamic area params. Params are same as in class constructor.

        :param kwargs: dynamic area params
        """
        vars(self).update(**kwargs)
        if 'data' in kwargs:
            self.width, self.height = self.data.size
            self.pagenum = self.data.page_count
            self.validlen = kwargs.get('validlen', self.width)
            self.dataoffset = kwargs.get('datalen', self.data.page_offset)

    def pixel_data(self):
        return bytes(self.data)

    def area_data(self, start_offset):
        pages_data = struct.pack('<h', self.pagenum)
        for i in range(self.pagenum):
            pages_data += struct.pack(
                '<I4BHBH9B2I',
                self.pagedatalen, self.pagestyle,
                self.displaymode, self.clearmode,
                self.speed, self.staytime, self.repeattime,
                self.validlen,
                *bytes(9),
                self.dataoffset * i + start_offset,
                self.dataoffset
            )

        if self.relateallpro:
            relateprogramms = struct.pack(
                '<BH',
                self.relateallpro,
                RESERVED
            )
        else:
            relateprogramms = struct.pack(
                '<BH%dH' % self.relatepronum,
                self.relateallpro,
                self.relatepronum,
                *self.relateproserials
            )

        return (
                struct.pack(
                    '<BH',
                    self.runmode, self.timeout,
                ) +
                relateprogramms +
                struct.pack(
                    '<BIB4HB',
                    self.overlap,
                    RESERVED,
                    self.areatype,
                    self.x, self.y,
                    self.width, self.height,
                    self.areaframe
                ) +
                bytes(9) +  # reserved 9 bytes
                pages_data
        )
