from typing import Tuple


def calc_padding(
        primary_size: Tuple[int, int],
        secondary_size: Tuple[int, int],
        align='center'
) -> Tuple[int, int]:
    """
    Calculates padding values when pasting images.

    Positioning:

    +-------------+--------+--------------+
    |`top-left`   |`top`   |`top-right`   |
    +-------------+--------+--------------+
    |`left`       |`center`|`right`       |
    +-------------+--------+--------------+
    |`bottom-left`|`bottom`|`bottom-right`|
    +-------------+--------+--------------+

    :param primary_size: primary image size
    :param secondary_size: pasting image size
    :param align: positioning
    """
    if not isinstance(align, str):
        align = 'center'
    px, py = map(
        lambda x: x[0] - x[1],
        zip(primary_size, secondary_size)
    )  # matches bottom-right
    alignment = align.split('-')
    if 'left' in alignment:
        px = 0
    elif 'center' in alignment or 'right' not in alignment:
        px = px // 2
    if 'top' in alignment:
        py = 0
    elif 'center' in alignment or 'bottom' not in alignment:
        py = py // 2
    return px, py


class PacketEncoder:
    """Packet encoding/decoding helper"""
    ESCAPE_DICT = [
        (b'\x5b', b'\x5b\x01'),
        (b'\x5a', b'\x5b\x02'),
        (b'\xa6', b'\xa6\x01'),
        (b'\xa5', b'\xa6\x02')
    ]

    @classmethod
    def encode(cls, data: bytes) -> bytes:
        for k, v in cls.ESCAPE_DICT:
            data = data.replace(k, v)
        return data

    @classmethod
    def decode(cls, data: bytes, unframe: bool = True) -> bytes:
        if unframe:
            data = cls.unframe_packet(data)
        for v, k in cls.ESCAPE_DICT:
            data = data.replace(k, v)
        return data

    @classmethod
    def frame_packet(cls, data: bytes, encode: bool = True):
        if encode:
            return b'\xa5' * 8 + cls.encode(data) + b'\x5a'
        else:
            return b'\xa5' * 8 + data + b'\x5a'

    @classmethod
    def unframe_packet(cls, data: bytes) -> bytes:
        return data.replace(b'\xa5', b'').replace(b'\x5a', b'')
